﻿/* 
    PASS Community Summit 2021
    Hacking SQL Server Another Day
    2021-11-10 (version 1.61)
    Demo - Hacking Encrypted Object Definitions
    André Melancia
    https://Andy.PT

    (C) 2016-2021, André Melancia

           ________   ________    _________  ____________;_
          - ______ \ - ______ \ / _____   //.  .  ._______/
         / /     / // /     / //_/     / // ___   /
        / /     / // /     / /       .-'//_/|_/,-'
       / /     / // /     / /     .-'.-'
      / /     / // /     / /     / /
     / /     / // /     / /     / /
    / /_____/ // /_____/ /     / /
    \________- \________-     /_/
         
*/

--  +----------------------------------------------------------------+
--  | MISSION BRIEFING                                               |
--  +----------------------------------------------------------------+

    -- WARNING: Run these scripts on an EMPTY database!

    -- INSTRUCTIONS: Run one statement at a time.
    

--  +----------------------------------------------------------------+
--  | MISSION PREPARATION                                            |
--  +----------------------------------------------------------------+

    USE AnotherDayAnotherHack; -- Or whatever database you created!
                               -- Please don't run this on the Master, 
                               -- that's just ugly!
    GO

    -- Things an MI6 agent with Licence to KILL has to do
    -- Note: "IF EXISTS" are ugly. First time these will generate errors
    --       Just accept the errors as part of an agent's life ;)
    DROP PROCEDURE dbo.TopSecretStoredProcedure_UnEncrypted;
	GO
    DROP PROCEDURE dbo.TopSecretStoredProcedure_ENCRYPTED;
	GO
	DROP VIEW dbo.TopSecretView_UnEncrypted;
	GO
	DROP VIEW dbo.TopSecretView_ENCRYPTED;
	GO


--  +----------------------------------------------------------------+
--  | CREATE Top secret UnEncripted stored procedure                 |
--  +----------------------------------------------------------------+

    GO
	CREATE PROCEDURE dbo.TopSecretStoredProcedure_UnEncrypted @SomeParameter NVarChar(255)
	AS
	Print 'Inside TopSecretStoredProcedure_UnEncrypted...';
	Print '    Parameter @SomeParameter = "' + @SomeParameter + '"...';
	Print 'Bye ;)';
	Print ' ';
	GO

	EXEC dbo.TopSecretStoredProcedure_UnEncrypted '1 trillion gold coins';
	GO

	EXEC SP_HelpText 'dbo.TopSecretStoredProcedure_UnEncrypted';
	GO


--  +----------------------------------------------------------------+
--  | CREATE Top secret ENCRIPTED stored procedure                   |
--  +----------------------------------------------------------------+

    GO
	CREATE PROCEDURE dbo.TopSecretStoredProcedure_ENCRYPTED @SomeParameter NVarChar(255)
	WITH ENCRYPTION -- <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	AS
	Print 'Inside TopSecretStoredProcedure_ENCRYPTED...';
	Print '    Parameter @SomeParameter = "' + @SomeParameter + '"...';
	Print '    ENCRYPTED TOP SECRET TEXT...';
	Print '    ENCRYPTED TOP SECRET TEXT...';
	Print '    ENCRYPTED TOP SECRET TEXT...';
	Print 'Bye ;)';
	Print ' ';
	GO

	EXEC dbo.TopSecretStoredProcedure_ENCRYPTED '1 trillion hidden gold coins';
	GO

	EXEC SP_HelpText 'dbo.TopSecretStoredProcedure_ENCRYPTED';
    GO


--  +----------------------------------------------------------------+
--  | CREATE Top secret UnEncripted view                             |
--  +----------------------------------------------------------------+

    GO
	CREATE VIEW dbo.TopSecretView_UnEncrypted 
	AS
	SELECT 'TopSecretView_UnEncrypted'                        AS WhateverNameThisIs,
	       'This is plain text for every enemy spy to see...' AS TopSecretMessage;
	GO

	SELECT *
	FROM   dbo.TopSecretView_UnEncrypted;

	GO
	EXEC SP_HelpText 'dbo.TopSecretView_UnEncrypted';
	GO


--  +----------------------------------------------------------------+
--  | CREATE Top secret ENCRIPTED view                               |
--  +----------------------------------------------------------------+

    GO
	CREATE VIEW dbo.TopSecretView_ENCRYPTED 
	WITH ENCRYPTION -- <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	AS
	SELECT 'TopSecretView_ENCRYPTED'                              AS WhateverNameThisIs,
	       'This is ENCRYPTED text for nobody to see...'          AS TopSecretMessage,
	       'ENCRYPTED'                                            AS Padding1,
		   'ENCRYPTED'                                            AS Padding2,
		   'ENCRYPTED'                                            AS Padding3;
	GO

	SELECT *
	FROM   dbo.TopSecretView_ENCRYPTED;

	GO
	EXEC SP_HelpText 'dbo.TopSecretView_ENCRYPTED';
	GO


--  +----------------------------------------------------------------+
--  | Functions and triggers...                                      |
--  +----------------------------------------------------------------+

    -- If you encrypt functions and triggers, these will work the same way!


--  +----------------------------------------------------------------+
--  | Hack every encrypted content the easy way                      |
--  +----------------------------------------------------------------+

	-- A lot of people have blogged on how to do this over the years, and
	-- many products exist that decrypt contents. Here is one:
	-- https://www.devart.com/dbforge/sql/sqldecryptor


--  +----------------------------------------------------------------+
--  | If you do decrypt in place...                                  |
--  +----------------------------------------------------------------+

	EXEC SP_HelpText 'dbo.TopSecretStoredProcedure_ENCRYPTED';
    GO

	EXEC SP_HelpText 'dbo.TopSecretView_ENCRYPTED';
	GO


--  +----------------------------------------------------------------+
--  | Mission is now over.                                           |
--  | You can't run away with the Bond Girl because you're an NSA    |
--  | agent, not an MI6 agent.                                       |
--  | Consider changing sides: MI6 has better... eh... cars...       |
--  +----------------------------------------------------------------+
