﻿/* 
    PASS Community Summit 2021
    Hacking SQL Server Another Day
    2021-11-10 (version 1.61)
    Demo - View security and Row-Level Security
    André Melancia
    https://Andy.PT

    (C) 2016-2021, André Melancia
    
    For more information on RLS please visit
    https://docs.microsoft.com/sql/relational-databases/security/row-level-security

           ________   ________    _________  ____________;_
          - ______ \ - ______ \ / _____   //.  .  ._______/
         / /     / // /     / //_/     / // ___   /
        / /     / // /     / /       .-'//_/|_/,-'
       / /     / // /     / /     .-'.-'
      / /     / // /     / /     / /
     / /     / // /     / /     / /
    / /_____/ // /_____/ /     / /
    \________- \________-     /_/
         
*/

--  +----------------------------------------------------------------+
--  | MISSION BRIEFING                                               |
--  +----------------------------------------------------------------+

    -- WARNING: Run these scripts on an EMPTY database!

    -- INSTRUCTIONS: Run one statement at a time.
    

--  +----------------------------------------------------------------+
--  | MISSION PREPARATION                                            |
--  +----------------------------------------------------------------+

    USE AnotherDayAnotherHack; -- Or whatever database you created!
                               -- Please don't run this on the Master, 
                               -- that's just ugly!
    GO

    -- Things an MI6 agent with Licence to KILL has to do
    -- Note: "IF EXISTS" are ugly. First time these will generate errors
    --       Just accept the errors as part of an agent's life ;)
    DROP USER M;
    DROP USER Q;
    DROP USER Sean;
    DROP USER David;
    DROP USER George;
    DROP USER Roger;
    DROP USER Timothy;
    DROP USER Pierce;
    DROP USER Daniel;
    GO
    
    DROP SECURITY POLICY AgentValidationPolicy
    GO
    DROP FUNCTION MI6_Access.FN_MI6_AgentValidationPredicate;
    GO
    DROP SCHEMA   MI6_Access;
    GO

    DROP VIEW     dbo.VIEW_BondFilms;
    GO
    DROP TABLE    dbo.BondFilms;
    GO
    DROP TABLE    dbo.Agents;
    GO
    DROP TABLE    dbo.Letters;
    GO
    

--  +----------------------------------------------------------------+
--  | CREATE USERS                                                   |
--  +----------------------------------------------------------------+

    CREATE USER M       WITHOUT LOGIN;
    CREATE USER Q       WITHOUT LOGIN;
    CREATE USER Sean    WITHOUT LOGIN;
    CREATE USER David   WITHOUT LOGIN;
    CREATE USER George  WITHOUT LOGIN;
    CREATE USER Roger   WITHOUT LOGIN;
    CREATE USER Timothy WITHOUT LOGIN;
    CREATE USER Pierce  WITHOUT LOGIN;
    CREATE USER Daniel  WITHOUT LOGIN;


--  +----------------------------------------------------------------+
--  | CREATE TABLE dbo.BondFilms                                     |
--  +----------------------------------------------------------------+

    CREATE TABLE dbo.BondFilms
    (
        FilmID       INT           NOT NULL,
        FilmTitle_EN NVARCHAR(64)  NOT NULL,
        FilmYear     INT           NOT NULL,
        BondActor    NVARCHAR(64)  NOT NULL,
        Agent        SYSNAME       NOT NULL,
		Payment      INT           NOT NULL,
        CONSTRAINT PK_BondFilms PRIMARY KEY CLUSTERED  ( FilmID ASC ) 
    );


    -- Insert some REAL top secret for-your-eyes-only information
    -- Leaked from here: https://en.wikipedia.org/wiki/List_of_James_Bond_films
    -- And fom here:     https://da.wikipedia.org/wiki/James_Bond
	INSERT INTO dbo.BondFilms
    (FilmID, FilmTitle_EN, FilmYear, BondActor, Agent, Payment) 
    VALUES 
    ( 1, 'Dr.No',                            1962, 'Sean Connery',   'Sean',      10000),
    ( 2, 'From Russia with Love',            1963, 'Sean Connery',   'Sean',      10000),
    ( 3, 'Goldfinger',                       1964, 'Sean Connery',   'Sean',      10000),
    ( 4, 'Thunderball',                      1965, 'Sean Connery',   'Sean',      10000),
    ( 5, 'Casino Royale',                    1967, 'David Niven',    'David',     10000),
    ( 6, 'You Only Live Twice',              1967, 'Sean Connery',   'Sean',      10000),
    ( 7, 'On Her Majesty''s Secret Service', 1969, 'George Lazenby', 'George',    10000),
    ( 8, 'Diamonds Are Forever',             1971, 'Sean Connery',   'Sean',      10000),
    ( 9, 'Live and Let Die',                 1973, 'Roger Moore',    'Roger',     10000),
    (10, 'The Man with the Golden Gun',      1974, 'Roger Moore',    'Roger',     10000),
    (11, 'The Spy Who Loved Me',             1977, 'Roger Moore',    'Roger',     10000),
    (12, 'Moonraker',                        1979, 'Roger Moore',    'Roger',     10000),
    (13, 'For Your Eyes Only',               1981, 'Roger Moore',    'Roger',     10000),
    (14, 'Octopussy',                        1983, 'Roger Moore',    'Roger',     10000),
    (15, 'Never Say Never Again',            1983, 'Sean Connery',   'Sean',    3552822), /* 3552822 = 0x363636 = "36 36 36 00" little endian */
    (16, 'A View to a Kill',                 1985, 'Roger Moore',    'Roger',     10000),
    (17, 'The Living Daylights',             1987, 'Timothy Dalton', 'Timothy',   10000),
    (18, 'Licence to Kill',                  1989, 'Timothy Dalton', 'Timothy',   10000),
    (19, 'GoldenEye',                        1995, 'Pierce Brosnan', 'Pierce',    10000),
    (20, 'Tomorrow Never Dies',              1997, 'Pierce Brosnan', 'Pierce',    10000),
    (21, 'The World Is Not Enough',          1999, 'Pierce Brosnan', 'Pierce',    10000),
    (22, 'Die Another Day',                  2002, 'Pierce Brosnan', 'Pierce',    10000),
    (23, 'Casino Royale',                    2006, 'Daniel Craig',   'Daniel',  3552822), /* 3552822 = 0x363636 = "36 36 36 00" little endian */
    (24, 'Quantum of Solace',                2008, 'Daniel Craig',   'Daniel',    10000),
    (25, 'Skyfall',                          2012, 'Daniel Craig',   'Daniel',    10000),
    (26, 'Spectre',                          2015, 'Daniel Craig',   'Daniel',    10000),
    (27, 'No Time To Die',                   2021, 'Daniel Craig',   'Daniel',    10000);


    -- Super advanced top secret spying technique
    SELECT * FROM dbo.BondFilms;

    GO


--  +----------------------------------------------------------------+
--  | VIEW BASED SECURITY DEMO                                       |
--  +----------------------------------------------------------------+

    CREATE VIEW dbo.VIEW_BondFilms
    WITH SCHEMABINDING
    AS
        SELECT FilmID, FilmTitle_EN, FilmYear, BondActor, Agent
        FROM   dbo.BondFilms
        WHERE  Agent = USER_NAME() OR USER_NAME() = 'M';
    GO


    -- We are so confident in our security that we're assigning read permissions
    --     to everyone.
    --     /!\ Please don't do this in a production environment, assign 
    --     permissions to the group(s) or role(s) the users belong
    GRANT SELECT ON dbo.VIEW_BondFilms TO Public;


    -- Now to test Q's tools!
    -- View the table as different agents
    
    EXECUTE AS USER = 'M';
    SELECT *, 'Seen as M' AS Who FROM dbo.VIEW_BondFilms; -- Returns all 26 missions
    REVERT; -- Go back to connection default user

    EXECUTE AS USER = 'Sean';
    SELECT *, 'Seen as Sean' AS Who FROM dbo.VIEW_BondFilms; -- Returns 7 missions
    REVERT;

    EXECUTE AS USER = 'Roger';
    SELECT *, 'Seen as Roger' AS Who FROM dbo.VIEW_BondFilms; -- Also returns 7 missions
    REVERT;

    EXECUTE AS USER = 'Q';
    SELECT *, 'Seen as Q' AS Who FROM dbo.VIEW_BondFilms; -- Returns nothing, "need to know" only!
    REVERT; 

    GO


--  +----------------------------------------------------------------+
--  | ROW-LEVEL SECURITY DEMO                                        |
--  +----------------------------------------------------------------+

    -- Assign "Licence to View" on the classified list "dbo.BondFilms" to agents
    GRANT SELECT ON dbo.BondFilms TO M;
    GRANT SELECT ON dbo.BondFilms TO Q;
    GRANT SELECT ON dbo.BondFilms TO Sean;   
    GRANT SELECT ON dbo.BondFilms TO David;
    GRANT SELECT ON dbo.BondFilms TO George;
    GRANT SELECT ON dbo.BondFilms TO Roger;
    GRANT SELECT ON dbo.BondFilms TO Timothy;
    GRANT SELECT ON dbo.BondFilms TO Pierce;
    GRANT SELECT ON dbo.BondFilms TO Daniel;
    GO
    -- It would be better to do this for a role, but this is just a demo!


    -- Create a new schema (RECOMMENDED best practice) 
    CREATE SCHEMA MI6_Access;
    GO


    -- Create an inline table valued function. 
    --     This function returns (1) when the Agent column is the same as the current
    --     user or if it's a top agent
    
    CREATE FUNCTION MI6_Access.FN_MI6_AgentValidationPredicate (@Agent AS SYSNAME)
    RETURNS TABLE
    WITH SCHEMABINDING
    AS
        RETURN SELECT 1 AS ValidationResult 
        WHERE @Agent = USER_NAME() OR USER_NAME() = 'M';
    GO


    -- Create a security policy using that function as filter predicate.
    --     Remember to set STATE=ON to enable it!
    CREATE SECURITY POLICY AgentValidationPolicy
    ADD FILTER PREDICATE MI6_Access.FN_MI6_AgentValidationPredicate (Agent)
    -- "ADD BLOCK PREDICATE" also exists
    ON dbo.BondFilms
    WITH (STATE = ON);
    GO

	-- Top secret spying technique to see if RLS exists
	SELECT OBJECT_NAME(object_id) AS Object_Name, * FROM sys.security_policies;
	SELECT OBJECT_NAME(object_id) AS Object_Name, * FROM sys.security_predicates;


    -- Now to test Q's tools!
    -- View the table as different agents
    
    EXECUTE AS USER = 'M';
    SELECT *, 'Seen as M using RLS' AS Who FROM dbo.BondFilms; -- Returns all 26 missions
    REVERT; -- Go back to connection default user

    EXECUTE AS USER = 'Sean';
    SELECT *, 'Seen as Sean using RLS' AS Who FROM dbo.BondFilms; -- Returns 7 missions
    REVERT;

    EXECUTE AS USER = 'Roger';
    SELECT *, 'Seen as Roger using RLS' AS Who FROM dbo.BondFilms; -- Also returns 7 missions
    REVERT;

    EXECUTE AS USER = 'Q';
    SELECT *, 'Seen as Q using RLS' AS Who FROM dbo.BondFilms; -- Returns nothing, "need to know" only!
    REVERT; 

    GO

    -- Bad bald cat-stroking guy killed, disable the security policy
    ALTER SECURITY POLICY AgentValidationPolicy
    WITH (STATE = OFF);
    GO

	-- Top secret spying technique to see if RLS exists
	SELECT OBJECT_NAME(object_id) AS Object_Name, * FROM sys.security_policies;
	SELECT OBJECT_NAME(object_id) AS Object_Name, * FROM sys.security_predicates;

    -- Now Q can see all missions
    EXECUTE AS USER = 'Q';
    SELECT *, 'Seen as Q after disabling RLS' AS Who FROM dbo.BondFilms; -- Returns everything now!
    REVERT; 
    
    GO


--  +----------------------------------------------------------------+
--  | ROW-LEVEL SECURITY EXPLOIT 1 - Knowing!                        |
--  +----------------------------------------------------------------+

    -- Re-enable security policy
    ALTER SECURITY POLICY AgentValidationPolicy
    WITH (STATE = ON);
    GO
	
    -- Am I being deceived by RLS?
    EXECUTE AS USER = 'Q';
	SELECT * FROM sys.partitions WHERE object_id = OBJECT_ID('dbo.BondFilms');
    REVERT; 
    -- Sys.Partitions permissions -> Public!
	GO

    
	
--  +----------------------------------------------------------------+
--  | ROW-LEVEL SECURITY EXPLOIT 2 - Old stuff                       |
--  +----------------------------------------------------------------+
	
	-- Bad bald cat-stroking guy killed, disable the security policy
    ALTER SECURITY POLICY AgentValidationPolicy
    WITH (STATE = OFF);
    GO


    -- No longer a security bug (or is it?)...
    EXECUTE AS USER = 'Q';
    SELECT *, 1 / (FilmYear - 1969), 'Seen as Q trying zero division - RLS OFF' AS Who FROM dbo.BondFilms; -- Returns error!
    REVERT; 




    -- Re-enable security policy
    ALTER SECURITY POLICY AgentValidationPolicy
    WITH (STATE = ON);
    GO
    
    EXECUTE AS USER = 'Q';
    SELECT *, 1 / (FilmYear - 1969), 'Seen as Q trying zero division - RLS ON' AS Who FROM dbo.BondFilms; -- Returns error!
    REVERT; 
    
    -- In the latest SQL Server 2016 CTPs and RTM, this is no longer a security issue.














--  +----------------------------------------------------------------+
--  | ROW-LEVEL SECURITY EXPLOIT 3 - Still working stuff             |
--  +----------------------------------------------------------------+
    
    -- Or is it? Muahahahahahah!
    EXECUTE AS USER = 'Q';
    SELECT *, 'Seen as Q trying zero division on WHERE - RLS ON' AS Who FROM dbo.BondFilms
    WHERE 666 = 1 / (FilmYear - 1969); -- Returns error!
    REVERT; 

    -- In SQL Server 2019 RTM or SQL Database, this is no longer a security issue.

--  +----------------------------------------------------------------+
--  | ROW-LEVEL SECURITY EXPLOIT 4 - AyRLSHacker tool                |
--  +----------------------------------------------------------------+
    
    -- Special thanks to 
    --     Scott Hanselman (Microsoft)
	--     @shanselman 


	-- AyRLSHacker tool
	-- (for security reasons that tool is currently unavailable
	--  until this bug has been fixed - thank you for understanding)
	


--  +----------------------------------------------------------------+
--  | Mission is now over.                                           |
--  | You can't run away with the Bond Girl because you're an NSA    |
--  | agent, not an MI6 agent.                                       |
--  | Consider changing sides: MI6 has better... eh... cars...       |
--  +----------------------------------------------------------------+
